;;; popper.el --- shrink-wrapped temporary windows for Emacs

;;;     Author: Chris McConnell <ccm@cs.cmu.edu>
;;; Maintainer: Greg Klanderman <greg.klanderman@alum.mit.edu>
;;;    Version: 2.40 XEmacs 1.0
;;;       Date: Feb 22, 1998
;;;    Created: 1990 (?)
;;;   Keywords: temporary window, popup window


;;; Copyright (C) 1990, 1991, 1992 Chris McConnell, ccm@cs.cmu.edu.
;;; Thanks to Ken Laprade for suggestions and patches.

;;; This file is part of GNU Emacs.

;;; GNU Emacs is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY.  No author or distributor
;;; accepts responsibility to anyone for the consequences of using it
;;; or for whether it serves any particular purpose or works at all,
;;; unless he says so in writing.  Refer to the GNU Emacs General Public
;;; License for full details.

;;; Everyone is granted permission to copy, modify and redistribute
;;; GNU Emacs, but only under the conditions described in the
;;; GNU Emacs General Public License.   A copy of this license is
;;; supposed to have been given to you along with GNU Emacs so you
;;; can know your rights and responsibilities.  It should be in a
;;; file named COPYING.  Among other things, the copyright notice
;;; and this notice must be preserved on all copies.


;;; Commentary:
;;; ~~~~~~~~~~~
;;; VERSION: This version of popper is based on version 2.40, which as
;;; far as I can tell is the last version distributed by Chris
;;; McConnell.  I have been unsuccessful at contacting him and so have
;;; decided to take over maintenance of the package for XEmacs.  New
;;; versions distributed by me will be designated "2.40 XEmacs X.Y".
;;; This version incorporates significant bug fixes and changes with
;;; respect to the original.  Overall it should work much better.  Let
;;; me know if you have any problems.  -- Greg Klanderman
;;;
;;;
;;; DESCRIPTION: This module provides a single shrink-wrapped window
;;; for displaying temporary text called the popper window.  At any
;;; time there is only one popper window on the screen.  If there is an
;;; entry for the buffer being displayed on popper-min-heights, the
;;; size of the window will be from that entry.  If the buffer is
;;; empty, the size of the window will be popper-empty-min.  Otherwise
;;; its size will be the minimum of the number of lines of text being
;;; displayed and half the number of lines in the currently selected
;;; window when the popper window was created.  It will work with any
;;; function that uses temporary windows or that has been wrapped with
;;; popper-wrap.  The window can be scrolled or buried from any other
;;; window.
;;;
;;; When a buffer is displayed using the function
;;; with-output-to-temp-buffer, the text will be displayed in the
;;; popper window if the name of the buffer is in popper-pop-buffers or
;;; popper-pop-buffers is set to `t' and the name is not in
;;; popper-no-pop-buffers.  If you have a buffer with a process, you
;;; can cause it to automatically scroll by setting
;;; popper-scroll-buffers to `t' or to a list of buffer names to
;;; scroll.  Many kinds of completion and help information are
;;; displayed this way.  In general any buffer with *'s around its name
;;; will be a temporary buffer.  Some commands like shell-command do
;;; not use with-output-to-temp-buffer even though you might like to
;;; have their output be temporary.  For commands like this, you can
;;; define a wrapper using the function popper-wrap.
;;;
;;; If popper-use-message-buffer is non-nil then popper output that is
;;; one line or less will be displayed in the minibuffer.
;;;
;;; The default binding for C-x o is changed so that when a buffer is
;;; displayed in a popper window, it will be skipped if it is in
;;; popper-buffers-to-skip or popper-buffers-to-skip is `t' and it is
;;; not in popper-buffers-no-skip.
;;;
;;; USAGE: Load this file, preferably after byte-compiling it.  Set the
;;; variable popper-pop-buffers to `t' to start using popper: by
;;; default popper does not enable itself by loading.  If you do not
;;; define key bindings using popper-load-hook, the bindings will be:
;;; 
;;;   * C-z C-z, C-z z, C-z 1 : popper-bury-output
;;;     Alternately hide or show the temporary pop-up window.
;;;   
;;;   * C-z C-v, C-z v : popper-scroll-output
;;;     Scroll the popper output window.
;;;   
;;;   * C-z C-g, C-z g : popper-grow-output
;;;     Grow the popper output window.
;;;   
;;;   * C-z C-b, C-z b : popper-switch
;;;     Prompt for a different popper temporary buffer to show.
;;;   
;;;   * C-z C-k, C-z k : popper-kill
;;;     Kill the current popper buffer.  If the stack of popper buffers 
;;;     is not empty, show the new topmost buffer.
;;;   
;;;   * C-z C-x, C-z x : popper-cycle
;;;     Cycle through all popper buffers.
;;;   
;;;   * C-z C-s, C-z s : popper-save
;;;     Save the current popper buffer by renaming it so that it won't be
;;;     overwritten. 
;;;   
;;;   * C-x o : popper-other-window
;;;     Like other-window, but skips the popper window.  Use a C-u prefix
;;;     to select the popper window.
;;;
;;; 
;;; See the section `Popper user variables' below for possible options.
;;; Here is a sample load hook for your .emacs:
;;;
;;; (setq popper-load-hook 
;;;      (lambda ()
;;;        (setq popper-prefix "\C-z")
;;;        (setq popper-pop-buffers t) ; Hooray!
;;;        (setq popper-buffers-to-skip popper-pop-buffers)
;;;        (setq popper-mode-line-text " Popper")
;;;        (setq popper-use-message-buffer nil)
;;;        (setq popper-no-pop-buffers '("^Man: "))
;;;        (setq popper-empty-min 1)
;;;        (setq popper-min-heights nil)
;;;        (setq popper-max-heights '((".*" . (40))))
;;;        (setq popper-delete-window-can-delete-frame nil)))
;;; (require 'popper)
;;;
;;; You may wrap a command in your popper-load-hook that does not use
;;; `with-output-to-temp-buffer' so that it will work with popper as
;;; follows:
;;;
;;; (popper-wrap command "*Command Output*")))
;;;
;;; This may not yet work in all cases.  It is known not to work for
;;; M-x compile at this time. (See TODO below)
;;;
;;; When loaded, popper checks the values of certain user variable for
;;; possible problems, unless popper-inhibit-warnings is non-nil.  In
;;; addition, unless popper-fix-incompatible-settings is nil, popper
;;; will set these variables to values that work with popper.  See also
;;; the variable popper-suggested-settings.
;;;
;;; WARNING: This package redefines the function split-window and
;;; pop-to-buffer so that the popper window is buried before calling
;;; the old definition.  The function popper-wrap uses defadvice.


;;; Change Log:
;;; ~~~~~~~~~~~
;;; Updated March 1997 -- February 1998 by Greg Klanderman
;;;
;;; fix up unreadable indentation 
;;; fix off-by-one error in popper-eat-newlines code
;;; fix lossage in popper-other-window
;;; modularize predicates for skip, handle, scroll
;;; took out fsf-19 alternate args for where-is-internal;
;;;      the docs didn't seem to indicate this was correct,
;;;      and it seg faulted FSF 19.27 when I tried it.
;;; much code cleanup and reorganization
;;; keep split-window from creating "half lines", 
;;;   eg, compute the size if the size arg is nil (meaning "split in half")
;;;   this caused popper windows to sometimes have a partial blank line or 
;;;   text line at the bottom which was rather annoying
;;;   this seems to also be the cause of the "won't change only window bug"
;;;   eg, the window is height 9, gets split into two windows of height 4
;;;   (actually 4.5, but window-height returns 4) then things get hosed...
;;;   also wrapped a (when more-than-one ..) around the enlarge-window
;;;   since the error may still happen when somebody mixes different
;;;   height fonts or has modeline-shadow-thickness non-zero.
;;; handle more emacs/XEmacs versions (hopefully! (let me know..))
;;; fix second off-by-one error in popper-shrink-window which allowed
;;;   it to inadvertently enlarge the window by at most one line
;;; added popper-max-heights variable to set max size of pop up windows
;;; add popper-split-preserve-popper-size, defaults to t, the old behavior
;;;   meant to deal with incongruity in behavior of split-window with and
;;;   without arg from the popper-parent window.
;;;   when set t: without arg, splits window in half as if popper weren't there
;;;   with arg N, splits at N lines, relative to bottom of pop up window.
;;;   when set nil: we hide just popper, split, and reshow popper without 
;;;   trying to preserve pop up size; its just like the popper weren't there
;;; found the variable split-window-keep-point, set nil = a good thing
;;; fix bug with split-window redefinition when called from split-window-vertically
;;;   and split-window-keep-point=nil.   is my solution too much of a hack??
;;; fix delete-window too (a la split-window)
;;;   instead of redefinition, define popper-delete-window and replace keymap entries
;;; delete-window deletes frame if its the only window - bad bad bad
;;;   popper-delete-window obeys variable popper-delete-window-can-delete-frame
;;; popper-delete-window avoids display jump when deleting topmost window
;;; popper-kill \C-z k and popper-cycle \C-z x to kill and cycle the popper buffer
;;; popper-save \C-z s to rename popper buffer so it gets saved
;;; introduce popper-prefix variable and popper-map keymap
;;; made the popper-pop-buffers list and kin be regexp lists


;;; TODO:
;;; ~~~~~
;;; don't ever let blank space be shown below (point-max)
;;;
;;; process stuff is broken
;;; process stuff should allow growing as output is produced
;;; figure out how to popper-wrap the byte compilation log
;;;
;;; restore window config if scrolled
;;;   we'll have to store the scroll amount, the buffer, and the line
;;;   and only scroll back if conditions are the same
;;;
;;; list ranking preference of window to split: 'current, 'topmost, 'biggest
;;;     current == current, or previous if minibuffer
;;;     others are self explanatory
;;;
;;; popper-scroll-back \C-z \M-v
;;; popper-split-preserve-popper-size - losing or not ??
;;;
;;; suggest setting
;;;   (setq help-selects-help-window nil
;;;         scroll-step 1
;;;         scroll-on-clipped-lines nil
;;;         pop-up-frames  nil
;;;         pop-up-windows t
;;;         temp-buffer-shrink-to-fit nil
;;;         split-window-keep-point nil)
;;; maybe other suggested variable values
;;;
;;; don't enable anything by default
;;; provide load hook to use, popper-setup-defaults or something similar?
;;;
;;; handle mult frames ???
;;; code cleanup
;;; fix up documentation

;;; ======================================================================


;;; Code:

;;;
;;; Popper user variables
;;;

(defvar popper-load-hook nil
  "*List of functions to run when the popper module is loaded.")

(defvar popper-prefix "\C-z"
  "*If non-nil, where to install the popper prefix keymap")

(defvar popper-pop-buffers nil
  "*List of buffers to put in the shrink-wrapped pop-up window.  
If it is T, all temporary buffers will be put in the pop-up window.")

(defvar popper-use-message-buffer nil
  "*If non-nil makes output to popper-buffers that is one line or less
  go to the minibuffer.")

(defvar popper-no-pop-buffers nil
  "*If popper-pop-buffers is T, these buffers will not be put into the
pop-up window.")

(defvar popper-buffers-to-skip popper-pop-buffers
  "*\\[popper-other-window] will skip over these buffers when they are
used in a temporary window.  If it is T, all popper windows will be
skipped except those in popper-buffers-no-skip.")

(defvar popper-buffers-no-skip nil
  "*\\[popper-other-window] will not skip these buffers when they are
used in a popper window if popper-buffers-to-skip is T.")

(defvar popper-eat-newlines t
  "*Boolean for eating trailing newlines in popper buffers.")

(defvar popper-scroll-buffers t
  "*If set to T or a list of buffer names, cause buffers with
associated processes to scroll automatically.")

(defvar popper-empty-min '(50) 
  "*Minimum number of lines to display for an empty popper buffer.
If it is a list, it is a percentage 0-100 of the available space.")

;;; #### does compilation even work??? -- No!
(defvar popper-min-heights '(("^\\*compilation\\*" . (50)))
  "*List of cons where each the car is a regular expression pattern to
match a buffer name and the cdr is the minimum number of lines to
allow when popping buffers that match the regular expression.  If the
number is a list, it is interpreted as the percentage of available
space 0-100 to use for the window.")

(defvar popper-max-heights nil
  "*List of cons where each the car is a regular expression pattern to
match a buffer name and the cdr is the maximum number of lines to
allow when popping up buffers that match the regular expression.  If the
number is a list, it is interpreted as the percentage of available
space 0-100 to use for the window.")

(defvar popper-mode-line-text nil
  "*Minor mode text for mode line of popper buffers.  If nil, it will
be set to a short help message on first use of popper.")

(defvar popper-inhibit-warnings nil
  "*When non-nil, popper checks the settings of certain Emacs user
variables that might adversely affect popper.")

(defvar popper-fix-incompatible-settings t
  "*When non-nil, if popper finds settings of certain Emacs user
variables that might adversely affect popper it sets them appropriately.")

(defvar popper-split-preserve-popper-size t
  "*Non-nil is split-window from the popper parent window should
preserve the pop up window's height.")

(defvar popper-delete-window-can-delete-frame t
  "*When non-nil, `popper-delete-window' with only one window
should delete the frame.")


;;;
;;; Popper internal variables
;;;

(defvar popper-output-buffers nil
  "LIFO list of buffers displayed in the popper window.")

(defvar popper-last-output-window nil
  "The window that last popped up an output window.")

(defvar popper-buffer nil
  "Indicates buffer is a popper for minor-mode-alist.")
(make-variable-buffer-local 'popper-buffer)

(defvar popper-original-filter nil
  "Original process filter, used by `popper-scroll-filter'.")
(make-variable-buffer-local 'popper-original-filter)

(defconst popper-temp-buffer-show-function
  (if (string-match "^18" emacs-version)
      'temp-buffer-show-hook                 ; older emacsen
    'temp-buffer-show-function)              ; newer GNU Emacs and XEmacs
  "Value is a symbol which names the variable holding the
temporary buffer show function for this emacs.  Value should
be either `temp-buffer-show-hook' or `temp-buffer-show-function',
depending on the version of emacs.")

(defvar popper-suggested-settings
  '((help-selects-help-window  . nil)
    (window-min-height         . 2)
    (scroll-step               . 1)
    (scroll-on-clipped-lines   . nil)
    (pop-up-frames             . nil)
    (pop-up-windows            . t)
    (temp-buffer-shrink-to-fit . nil)
    (split-window-keep-point   . nil))
  "Suggested settings for Emacs variables when popper is in use.")


;;;
;;; General purpose utilities - sure would be nice if these were in emacs
;;;

(defun popper-mem (item list &optional elt=)
  "Test to see if ITEM is equal to an item in LIST.
Optional comparison function ELT= defaults to equal."
  (let ((elt= (or elt= (function equal)))
	(done nil))
    (while (and list (not done))
      (if (funcall elt= item (car list))
	  (setq done list)
        (setq list (cdr list))))
    done))

(defun popper-assoc (item list &optional elt=)
  "Test to see if ITEM is equal to the car of an element of LIST.
Optional comparison function ELT= defaults to equal.
The value is actually the element of LIST whose car equals KEY."
  (let ((elt= (or elt= (function equal)))
	(done nil))
    (while (and list (not done))
      (if (funcall elt= item (car (car list)))
	  (setq done (car list))
        (setq list (cdr list))))
    done))

(defun popper-match-buffer-name (buff regx)
  "Match a buffer name to a regular expression"
  (string-match regx buff))

(defun popper-filter (list &optional pred)
  "Filter LIST with predicate PRED.
Return a new list containing all elements in LIST for which PRED returns
a non-nil value."
  (let ((pred (or pred (function (lambda (x) x))))
        (result nil))
    (while list
      (if (funcall pred (car list))
          (setq result (cons (car list) result)))
      (setq list (cdr list)))
    (nreverse result)))


;;;
;;; Popper internal utilities 
;;;

(defun popper-warn (&rest args)
  "Display a popper warning."
  (display-warning 'popper (apply 'format args)))

(defun popper-handle-buffer-p (&optional buff)
  "Determine if popper is handling BUFF.  BUFF can be a buffer name
or a buffer.  If not specified, it defaults to the current buffer."
  (let ((buff-name  (if (stringp buff) buff (buffer-name buff))))
    (or (and (eq popper-pop-buffers t)
             (not (popper-mem buff-name popper-no-pop-buffers
                              'popper-match-buffer-name)))
        (and (listp popper-pop-buffers)
             (popper-mem buff-name popper-pop-buffers
                         'popper-match-buffer-name)))))

(defun popper-scroll-buffer-p (&optional buff)
  "Determine if popper should scroll process output in BUFF"
  (let ((buff-name  (if (stringp buff) buff (buffer-name buff))))
    (or (eq popper-scroll-buffers t)
        (popper-mem buff-name popper-scroll-buffers
                    'popper-match-buffer-name))))

(defun popper-skip-buffer-p ()
  "Determine if popper-other-window should skip the current buffer"
  (let* ((buff (current-buffer))
         (name (buffer-name buff)))
    (or (and (eq popper-buffers-to-skip t)
             (not (popper-mem name popper-buffers-no-skip
                              'popper-match-buffer-name))
             (eq (popper-output-buffer) buff))
        (and (listp popper-buffers-to-skip)
             (popper-mem buff popper-buffers-to-skip
                         'popper-match-buffer-name)))))

(defun popper-select (&optional window)
  "Select WINDOW and its buffer.  WINDOW defaults to selected-window."
  (setq window (or window (selected-window)))
  (select-window window)
  (set-buffer (window-buffer window)))

(defun popper-first-buffer ()
  "Remove killed buffers and return the first buffer on
popper-output-buffers."
  (setq popper-output-buffers
        (popper-filter popper-output-buffers
                       (lambda (buf) (buffer-name buf))))
  (car popper-output-buffers))

(defun popper-output-buffer ()
  "Return the buffer being displayed in the popper window."
  (popper-first-buffer)
  (if (not (eq (selected-window) (next-window)))
      (let ((buffers popper-output-buffers)
	    (done nil))
	(while buffers
	  (let* ((buffer (car buffers))
		 (window (if (buffer-name buffer) (get-buffer-window buffer))))
	    (if (and window (not done))
		(setq done buffer)
              (save-excursion
                (set-buffer (car buffers)) (setq popper-buffer nil))
              (setq buffers (cdr buffers)))))
	(if done
            ;; move it to front
            (setq popper-output-buffers
                  (cons done (delq done popper-output-buffers))))
	done)))

(defun popper-parent ()
  "Return the parent of the popper window."
  (let ((output (popper-output-buffer)))
    (if output (next-window (get-buffer-window output) 'no))))

(defun popper-window-heights (window)
  "Return a list of the heights of all of the windows following WINDOW."
  (let ((heights nil))
    (select-window window)
    (while (progn 
	     (select-window (next-window (selected-window) 'no))
	     (not (eq (selected-window) window)))
      (setq heights (cons (window-height (selected-window)) heights)))
    (nreverse heights)))

(defun popper-min-height ()
  "Return the minimum height to use for the popper buffer in the
current window.  This is either an entry from popper-min-heights,
popper-empty-min if the buffer is empty or window-min-height."
  (let ((min (or (cdr (popper-assoc (buffer-name) popper-min-heights
                                    'popper-match-buffer-name))
                 (if (= (point-min) (point-max))
                     popper-empty-min
                   window-min-height))))
    (if (consp min)
	;; Floating percentage
	(round (/ (* (+ (window-height) (window-height (next-window)))
                     (car min))
                  100.0))
      min)))

(defun popper-max-height (buff)
  "Return the maximum height to use for a popper buffer to be
created by splitting the current window.  This is either an entry
from popper-max-heights, or half the height of the current window."
  (let ((val (or (cdr (popper-assoc (buffer-name buff) popper-max-heights
                                    'popper-match-buffer-name))
                 '(50))))
    (when (consp val)                        ; Floating percentage
	(setq val (round (/ (* (window-height) (car val)) 100.0))))
    (max (min val (- (window-height) window-min-height))
         window-min-height)))

;;; #### look at this
(defun popper-scroll-filter (process output)
  "Scroll and keep last point in window."
  (let* ((old (selected-window))
	 (buffer (process-buffer process))
	 (window (get-buffer-window buffer))
	 ;*w ;;; #### this was unused except for the assignment below ??
         )
    (save-excursion
      (set-buffer buffer)
      (if popper-original-filter
	  (funcall popper-original-filter process output)
        (goto-char (process-mark process))
        (insert output)
        (set-marker (process-mark process) (point))))
    (if (and window)
	(progn
	  ;(setq *w window)
	  (select-window window)
	  (goto-char (point-max))
	  (move-to-window-line nil)
	  (select-window old)))))

;;; #### probably other cleanup
(defun popper-show-output (&optional buffer size)
  "Bring the output window up showing optional BUFFER in window of
SIZE.  If SIZE is not specified, then shrink the window.  Finally
select the original window."
  (let* ((window (selected-window))
	 (old-buffer (window-buffer window))
	 (buffer (get-buffer-create
		  (or buffer (popper-first-buffer) 
		      (error "No popper buffers"))))
	 start parent
	 (min-height (+ window-min-height (or size window-min-height)))
	 (text-lines (save-excursion
		       (set-buffer buffer)
		       (count-lines (point-min) (point-max)))))
    (if (and (= text-lines 1) popper-use-message-buffer)
	(message 
	 (save-excursion
	   (set-buffer buffer) (buffer-substring (point-min) (point-max))))
      (setq popper-last-output-window window)
      (if (eq buffer old-buffer)
          (popper-shrink-window)
	(if (eq window (minibuffer-window)) 
	    (let* ((parent (popper-parent)))
	      (popper-bury-output t)
	      (select-window (setq window (or parent (previous-window)))))
          (if (not (eq old-buffer (popper-output-buffer)))
              (popper-bury-output t)))
	(if (< (window-height window) min-height)
	    (enlarge-window (- min-height (window-height window))))  ;; #### ok ?? 
	(setq start (window-start window))
	(split-window nil (or size (popper-max-height buffer)))
	(set-window-buffer window buffer)
	(set-buffer buffer)
	(setq popper-buffer t)
	(setq popper-output-buffers
	      (cons buffer (delq buffer popper-output-buffers)))
	(if (not size) (popper-shrink-window))
	(setq parent (next-window window 'no))
	(popper-select parent)
	;; Move the window so that top lines get covered unless it would
	;; cover point in which case point is at top of window
        ;; #### figure this out and save how you scrolled so we can
        ;; #### scroll back when we later bury the popper window
	(save-excursion
	  (set-window-start parent start)
	  (move-to-window-line (window-height window))
	  (set-window-start parent (point)))
	(let ((point (save-excursion (beginning-of-line) (point))))
	  (if (not (pos-visible-in-window-p point))
	      (set-window-start (selected-window) point)))
	(if (eq popper-last-output-window (minibuffer-window))
	    (select-window (minibuffer-window)))
	(set-buffer old-buffer)))))

;;; #### hacked a bit already, could prob use more
;;; #### process stuff needs testing
;;; #### be careful not to allow blank space shown after (point-max)
(defun popper-shrink-window ()
  "Shrink the current window if larger than its buffer unless it has
an entry in popper-min-heights or it is empty in which case
popper-empty-min is used."
  (let* ((window (selected-window))
	 (window-lines (1- (window-height window))))
    (set-buffer (window-buffer window))
    (let ((process (get-buffer-process (current-buffer))))
      ;; fix up leading and trailing space if desired and buffer has no process
      (when (and popper-eat-newlines (not process))
        (let ((buffer-read-only nil)
              (buffer-modified-p (buffer-modified-p)))
          (save-excursion
            ;; Delete trailing blank space
            (goto-char (point-max))
            (skip-chars-backward " \t\n")
            (if (< (point) (point-max))
                (delete-region (point) (point-max)))
            ;; Delete leading blank space
            (goto-char (point-min))
            (skip-chars-forward " \t\n")
            (beginning-of-line)
            (if (> (point) (point-min))
                (delete-region (point-min) (point)))
            ;; restore modified-p
            (set-buffer-modified-p buffer-modified-p))))
      ;; keep scrolling process pop-ups
      (when (and process (popper-scroll-buffer-p (current-buffer)))
        (while (not (marker-position (process-mark process)))
          (accept-process-output))
        (goto-char (point-max))
        ;; (set-marker (process-mark process) (point))
        (insert ? )
        (set-mark (point))
        (when (not popper-original-filter)
          (setq popper-original-filter (process-filter process))
          (set-process-filter process 'popper-scroll-filter)))
      ;; finally, shrink as needed
      (enlarge-window (- (max (1+ (save-excursion 
                                    (goto-char (point-min))
                                    (vertical-motion (1- window-lines))))
                              (1- (popper-min-height)))
                         window-lines)))))

(defun popper-show (buffer)
  "Function to display BUFFER in a popper window if it is in
popper-pop-buffers or popper-pop-buffers is T and it is not in
popper-no-pop-buffers.  This is used by popper as the value of
`temp-buffer-show-function'."
  (if (popper-handle-buffer-p buffer)
      (popper-show-output buffer)
    (display-buffer buffer))
  (setq minibuffer-scroll-window (get-buffer-window buffer)))

(defun popper-read-popper-buffer (prompt &optional default)
  "Read the name of a popper buffer and return as a string.
Prompts with PROMPT.  Optional second arg DEFAULT is value
to return if user enters an empty line."
  (let* ((prompt (if default 
                     (format "%s(default %s) "
                             (gettext prompt) (if (bufferp default)
                                                  (buffer-name default)
                                                default))
                   prompt))
         (alist (mapcar (lambda (b) (cons (buffer-name b) b))
                        popper-output-buffers))
         (result (completing-read prompt alist nil t
                                  nil 'buffer-history)))
    (if (equal result "")
        (setq result default))
    (if (bufferp result)
        (buffer-name result)
      result)))


;;;
;;; Popper commands
;;;

;;; #### looks good - should roll back if needed
(defun popper-bury-output (&optional no-show)
  "Bury the popper output.  If no popper output window is active, show 
the popper output window, unless optional argument NO-SHOW is non-nil."
  (interactive)
  (let ((buffer (popper-output-buffer)))
    (if buffer
	(let* ((old (current-buffer))
	       (old-window (selected-window))
	       (output (get-buffer-window buffer))
	       (start (window-height output))
	       (parent (next-window output 'no))
	       (height (window-height output))
	       (heights (popper-window-heights output))
               (more-than-one (cdr heights)))
	  (bury-buffer buffer)
	  (delete-window output)
	  (popper-select parent)
          (while heights
            ;; only resize if it makes sense, eg there's more than one window
            (when more-than-one
              (enlarge-window (- (+ height (car heights)) (window-height))))
            ;; #### this needs to roll back if necessary...
            (if start
                (condition-case () (scroll-down start) (error nil)))
            (select-window (next-window (selected-window) 'no))
            (setq height 0 start nil)
            (setq heights (cdr heights)))
	  (set-buffer old)
	  (if (not (eq old-window output))
	      (select-window old-window)))
      (if (not no-show) (popper-show-output)))))

;;; #### look at this
;;; #### use C-u prefix arg to  scroll back a page?
;;; #### don't allow blank space after end of buffer
(defun popper-scroll-output (&optional n)
  "Scroll text of the popper window upward ARG lines ; or near full
screen if no ARG.  When calling from a program, supply a number as
argument or nil.  If the output window is not being displayed, it will
be brought up."
  (interactive "P")
  (let ((buffer (popper-output-buffer)))
    (if buffer
	(let ((window (selected-window)))
	  (unwind-protect
              (progn (select-window (get-buffer-window buffer))
                     (condition-case ()
                         (scroll-up n)
                       (error
                        (if (or (null n) (and (numberp n) (> n 0)))
                            (set-window-start (selected-window) (point-min))
                          (while (not (pos-visible-in-window-p (point-max)))
                            (scroll-up nil))))))
	    (select-window window)))
      (popper-show-output))))

;;; #### look at this
(defun popper-grow-output (&optional n)
  "Grow the popper window by ARG (default 1) lines.  If the popper
window is not being shown, it will be brought up."
  (interactive "p")
  (let ((buffer (popper-output-buffer)))
    (if buffer
	(let ((old-buffer (current-buffer))
	      (window (selected-window)))
	  (select-window (get-buffer-window buffer))
	  (enlarge-window n)
	  (popper-select (next-window (selected-window) 'no))
	  (save-excursion
	    (if (< n 0)
		(condition-case () (scroll-up n) (error nil))
              (move-to-window-line n)
              (set-window-start (selected-window) (point))))
	  (select-window window)
	  (set-buffer old-buffer))
      (popper-show-output))))

(defun popper-other-window (arg)
  "Select the arg'th other window.  If arg is a C-u prefix, the popper
window will be selected.  Otherwise, windows that contain buffers in
popper-buffers-to-skip will be skipped or if popper-buffers-to-skip is
T those that are not in popper-buffers-no-skip.  This is the popper
 replacement for the function `other-window'"
  (interactive "P")
  (if (consp arg)
      ;; select popper window
      (let* ((buffer (popper-output-buffer))
	     (window (and buffer (get-buffer-window buffer))))
	(when window (select-window window)))
    ;; otherwise, normal other-window behavior, but skip popper windows
    (let (dirn)
      (setq arg  (if (eq arg '-) -1 (if (integerp arg) arg 1)))
      (setq dirn (signum arg))
      (setq arg  (abs arg))
      (while (> arg 0) 
        (other-window dirn)
        ;; if we landed on a buffer we want to skip, keep going
        ;; until we find one which is ok or we exhaust all windows
        (when (popper-skip-buffer-p)
          (let ((last-win (previous-window (selected-window))))
            (while (and (popper-skip-buffer-p)
                        (not (eq (selected-window) last-win)))
              (other-window dirn))))
        ;; decr arg
        (setq arg (1- arg))))))

(defun popper-switch (buffer)
  "Switch the popper window to BUFFER.  Interactively,
prompt for another popper output buffer to switch to."
  (interactive
   (list (let ((dflt (car (if (popper-output-buffer)
                              (cdr popper-output-buffers)
                            popper-output-buffers))))
           (when dflt 
             (popper-read-popper-buffer "Popper buffer " dflt)))))
  (if buffer
      (popper-show buffer)
    (error "No popper buffer to switch to")))

(defun popper-kill ()
  "Remove the current popper buffer from the LIFO stack of popper 
buffers and show the new topmost buffer, if there is one."
  (interactive)
  (let ((buf (popper-output-buffer)))
    (when buf
      (popper-bury-output t)
      (setq popper-output-buffers (delq buf popper-output-buffers))
      (if (string-match "^popper-save:" (buffer-name buf))
          (kill-buffer buf)
        (bury-buffer buf))
      (if (car popper-output-buffers)
          (popper-show-output)))))

(defun popper-cycle ()
  "Move the current popper buffer to the end of the LIFO stack of 
popper buffers and show the new topmost buffer, if there is one."
  (interactive)
  (let ((buf (popper-output-buffer)))
    (when buf
      (popper-bury-output t)
      (setq popper-output-buffers
            (append (delq buf popper-output-buffers)
                    (list buf)))
      (if (car popper-output-buffers)
          (popper-show-output)))))

(defun popper-save ()
  "Rename the current popper buffer to a unique name,
so that it will be preserved."
  (interactive)
  (let ((buf (popper-output-buffer)))
    (when buf
      (save-excursion
        (set-buffer buf)
        (rename-buffer
         (generate-new-buffer-name
          (concat "popper-save:" (buffer-name buf))))))))

;;; #### need a variable to control delete-frame behavior
;;; #### if we delete the frame, how to pick new popper frame ??
;;; #### also deleting top window munches display -
;;; #### need hack like in split-window-vertically to reduce display jumps
;;; #### restore popper parent too
(defun popper-do-delete-window (window force)
  "Helper function for `popper-delete-window'."
  (or window (setq window (selected-window)))
  (if (or popper-delete-window-can-delete-frame
          (cdr (window-list nil 'never)))
      (let* ((sib (next-window window 'never))
             (fixup (and (eq window (frame-highest-window))
                         (not (eq window sib))))
             (dwin-ht (window-height window)))
        ;; delete it
        (delete-window window force)
        ;; when deleting the topmost window, counteract the display jump
        (when fixup
          (save-excursion
            (set-buffer (window-buffer sib))
            (goto-char (window-start sib))
            (vertical-motion (- dwin-ht))
            (set-window-start sib (point)))))
    (message "can't delete last window on a frame")))

(defun popper-delete-window (&optional window force)
  "Remove WINDOW from the display.  Default is selected window.
If window is the only one on the frame, the frame is destroyed.
Normally, you cannot delete the last non-minibuffer-only frame (you must
use `save-buffers-kill-emacs' or `kill-emacs').  However, if optional
second argument FORCE is non-nil, you can delete the last frame. (This
will automatically call `save-buffers-kill-emacs'.)

This function is similar to `delete-window', but buries the popper window
when the window is actually being deleted.  Also, if the variable 
`popper-delete-window-can-delete-frame' is nil, the frame won't be deleted
if the window is the only window on the frame.  Finally, when the topmost
window is deleted, the display jump is avoided if possible."
  (interactive)
  (let ((parent (popper-parent)))
    (if parent
	(let ((pop-size (and popper-split-preserve-popper-size
                              (window-height
                               (get-buffer-window (popper-output-buffer))))))
	  (popper-bury-output t)
	  (prog1
	      (popper-do-delete-window window force)
            (popper-show-output nil pop-size)))  ;; #### pop-size losing ??
      (popper-do-delete-window window force))))


;;;
;;; Function redefinitions
;;;

;;; #### check this out -- can/should these use advice ???
;;; *** Redefine split-window to bury popper window first ***
(defvar popper-split-window (symbol-function 'split-window) 
  "Original definition of split-window.")

(defun popper-do-split-window (&optional window size hor-flag)
  "Helper function for popper redefinition of split-window"
  ;; compute the size if not given -- 
  ;; prevents a line from getting split in half
  ;; which is ugly and causes other nightmares because
  ;; then the sum of the new window-heights is not equal to
  ;; the original window-height
  (unless size
    (setq size
          (max (/ (if hor-flag (window-width window) (window-height window)) 2)
               (if hor-flag window-min-width window-min-height))))
  (funcall popper-split-window window size hor-flag))

;;; #### hack; should we restore the popper-parent too?
(defvar popper-split-window-pop-size nil)
(defun popper-split-window-reshow-output ()
  (remove-hook 'post-command-hook 'popper-split-window-reshow-output)
  (popper-show-output nil popper-split-window-pop-size)
  (setq popper-split-window-pop-size nil))

(defun split-window (&optional window size hor-flag)
  "Split WINDOW, putting SIZE lines in the first of the pair.
WINDOW defaults to selected one and SIZE to half its size.
If optional third arg HOR-FLAG is non-nil, split side by side
and put SIZE columns in the first of the pair."
  (let ((parent (popper-parent)))
    (if (eq parent (selected-window))
	(let ((pop-size (and popper-split-preserve-popper-size
                              (window-height
                               (get-buffer-window (popper-output-buffer))))))
          (if (and size pop-size (not hor-flag))
              (setq size (+ size pop-size)))
	  (popper-bury-output t)
	  (prog1
	      (popper-do-split-window window size hor-flag)
	    (if (and (eq this-command 'split-window-vertically)
                     (not split-window-keep-point))
                (progn ; #### total hack to reshow *after* s-w-vertically
                  (setq popper-split-window-pop-size pop-size)
                  (add-hook 'post-command-hook 'popper-split-window-reshow-output)
                  )
              (popper-show-output nil pop-size))))  ;; #### pop-size losing ??
      (popper-do-split-window window size hor-flag))))

;;; #### check this out - use advice?
;;; *** Redefine pop-to-buffer to skip output windows ***
(defvar popper-pop-to-buffer (symbol-function 'pop-to-buffer)
  "Original pop to buffer function.")
(defun pop-to-buffer (buffer &optional other-window on-frame)
  "Select buffer BUFFER in some window, preferably a different one.
If pop-up-windows is non-nil, windows can be split to do this.
If second arg OTHER-WINDOW is non-nil, insist on finding another
window even if BUFFER is already visible in the selected window."
  (let ((parent (popper-parent)))
    (if (and parent 
	     (not (eq (get-buffer buffer) (popper-output-buffer))))
	(progn
	  (popper-bury-output t)
	  (funcall popper-pop-to-buffer buffer other-window on-frame)
	  (select-window parent)
	  (popper-show-output)
	  (sit-for 0)			;Allow display update
	  (funcall popper-pop-to-buffer buffer nil on-frame)) ;; #### why ??
      (funcall popper-pop-to-buffer buffer other-window on-frame))))


;;;
;;; Wrap functions that don't use with-output-to-temp-buffer
;;;

;;; #### do these work? - yes
;;; No.  Don't do this when bytecompiling.
;;;   It is also evil to change XEmacs state (especially with advice)
;;;   when loading. -slb
;;; #### make it work for compile, grep, bytecompile, display-warning-buffer

(defvar popper-installed nil)

(defun popper-install ()
  (unless popper-installed
    (setq popper-installed t)
;;; #### move toward top?
    (require 'advice)
    (ad-start-advice)

;;; #### check this out 
    (defmacro popper-wrap (function buffer)
      "Define a wrapper on FUNCTION so that BUFFER will be a pop up window."
      (let ((name (intern (concat "popper-wrapper-" (symbol-name function)))))
	`(defadvice ,function (around ,name activate)
	   (let ((shown nil))
	     (save-window-excursion 
	       ad-do-it
	       (setq shown (get-buffer-window ,buffer)))
	     (if shown
		 (funcall ,popper-temp-buffer-show-function ,buffer))))))

    (popper-wrap shell-command           "*Shell Command Output*")
    (popper-wrap shell-command-on-region "*Shell Command Output*")

;;;
;;; perform installation actions
;;;

;; install temp-buffer-show-function/temp-buffer-show-hook
;; This also doesn't work when bytecompiling. -slb
    (set popper-temp-buffer-show-function 'popper-show)))

;; run our load-hook
(run-hooks 'popper-load-hook)

;; set default key bindings unless already specified by user
(defvar popper-map nil "Popper keymap")
(unless popper-map
  ;; make the keymap.  user can override completely by making
  ;; their own map from the load hook
  (setq popper-map (make-sparse-keymap "popper-map"))

  ;; define standard bindings
  (define-key popper-map "1"    'popper-bury-output)
  (define-key popper-map "z"    'popper-bury-output)
  (define-key popper-map "\C-z" 'popper-bury-output)

  (define-key popper-map "v"    'popper-scroll-output)
  (define-key popper-map "\C-v" 'popper-scroll-output)

  (define-key popper-map "g"    'popper-grow-output)
  (define-key popper-map "\C-g" 'popper-grow-output)

  (define-key popper-map "b"    'popper-switch)
  (define-key popper-map "\C-b" 'popper-switch)

  (define-key popper-map "k"    'popper-kill)
  (define-key popper-map "\C-k" 'popper-kill)

  (define-key popper-map "x"    'popper-cycle)
  (define-key popper-map "\C-x" 'popper-cycle)

  (define-key popper-map "s"    'popper-save)
  (define-key popper-map "\C-s" 'popper-save)
  )

;; install the popper-map
(when popper-prefix
  (define-key global-map popper-prefix popper-map))

;; replace bindings for `other-window' with `popper-other-window'
(substitute-key-definition 'other-window
                           'popper-other-window
                           global-map)

;; replace bindings for `delete-window' with `popper-delete-window'
(substitute-key-definition 'delete-window
                           'popper-delete-window
                           global-map)

;; install modeline indication for popper
(unless (assq 'popper-buffer minor-mode-alist)
  (setq minor-mode-alist
        (cons '(popper-buffer popper-mode-line-text) minor-mode-alist)))

;; set popper modeline text unless already specified by user
;; do this *after* keys get set...
(unless popper-mode-line-text
  (setq popper-mode-line-text
        (list (format 
               " %s bury, %s scroll" 
               (key-description
                (where-is-internal 'popper-bury-output nil t))
               (key-description
                (where-is-internal 'popper-scroll-output nil t))))))


;;;
;;; warn about possible trouble...
;;;
(unless popper-inhibit-warnings
  (mapc (lambda (pair)
          (when (boundp (car pair))
            (unless (eq (eval (car pair)) (cdr pair))
              (popper-warn "Popper suggests setting variable `%s' to %s.%s"
                           (car pair) (cdr pair)
                           (if popper-fix-incompatible-settings " Did it." ""))
              (if popper-fix-incompatible-settings
                  (set (car pair) (cdr pair))))))
        popper-suggested-settings)
  (setq popper-inhibit-warnings t))


;;;
;;; finally provide the feature
;;;

(provide 'popper)


;;; popper.el ends here
